use std::collections::HashMap;

use regex::Regex;
use trees::{tr, Node, Tree};

#[derive(Debug, Clone, Copy)]
pub enum Operation {
    Add,
    Sub,
    Mul,
    Div,
}

impl Operation {
    pub fn from(s: &str) -> Operation {
        match s {
            "+" => Operation::Add,
            "-" => Operation::Sub,
            "*" => Operation::Mul,
            "/" => Operation::Div,
            _ => panic!("Unknown operation {}", s),
        }
    }
}

#[derive(Debug, Clone)]
pub enum Monkey {
    Number(isize),
    Calculate(String, String, Operation),
}

pub fn parse_monkeys(input: &String) -> Tree<Monkey> {
    let mut monkeys = HashMap::new();

    let re = Regex::new(r"^(\w+): (?:(\d+)|(\w+) ([+\-*/]) (\w+))$").unwrap();
    for line in input.lines() {
        let captures = re.captures(line).unwrap();
        let name = captures.get(1).unwrap().as_str();
        let monkey = if let Some(num) = captures.get(2) {
            Monkey::Number(num.as_str().parse().unwrap())
        } else {
            Monkey::Calculate(
                captures.get(3).unwrap().as_str().to_owned(),
                captures.get(5).unwrap().as_str().to_owned(),
                Operation::from(captures.get(4).unwrap().as_str()),
            )
        };
        monkeys.insert(name.to_owned(), monkey);
    }

    let tree = to_tree(&monkeys, "root");

    tree
}

fn to_tree(monkeys: &HashMap<String, Monkey>, monkey: &str) -> Tree<Monkey> {
    let monkey = &monkeys[monkey];
    let mut tree = tr(monkey.clone());

    if let Monkey::Calculate(a, b, _) = monkey {
        let forest = -to_tree(monkeys, a) - to_tree(monkeys, b);
        tree.append(forest);
    }

    tree
}

pub fn yell(monkey: &Node<Monkey>) -> isize {
    match monkey.data() {
        Monkey::Number(n) => *n,
        Monkey::Calculate(_, _, op) => {
            let mut children = monkey.iter();
            let a = children.next().unwrap();
            let b = children.next().unwrap();
            match op {
                Operation::Add => yell(a) + yell(b),
                Operation::Sub => yell(a) - yell(b),
                Operation::Mul => yell(a) * yell(b),
                Operation::Div => yell(a) / yell(b),
            }
        }
    }
}
